#include "vector.cpp"

vector<vec4> cell;   // 600-cell before rotation

vector<vec4> cell2;  // 600-cell rotated

vector<vec3> vpos;   // positions of vertices

int edge[120][120];  // edge[i][j] > 0 iff cell[i]-cell[j] is an edge;
                     // the specific value denotes the verification status
                     // for this edge (1-white, 2-red, 3-green)

int vert = 120;      // the number of vertices (including the four
                     // vertices of the 600-cell which are unused in
                     // the triangulation)

int  vtype[120];     // vertex type; vertices are divided into layers
                     // numbered from 0 (four unused vertices) to 12
                     // (the most internal layer)

bool movit[120];     // movit[i] iff vertex i selected

int faces[600*6][4]; // for each 3-simplex, we give indices of all its
                     // vertices; for simplicity each 3-simplex is given
                     // in 6 permutations, so that each edge appears as
                     // faces[i][0]-faces[i][1] for some i
int nf;              // how many faces are there

int flat;            // the external vertex (of type 0) which is
                     // used as a direction of flattening

int corn[4];         // indices of the four corners of the tetrahedron (i.e.
                     // vertices type 1); corn[3] is opposite flat

int redtri[12];      // red vertices (i.e. vertices of type 3),
                     // in groups of three; 9 to 11 are on the face which
                     // remains regular

#include "graph.cpp"

// generate and rotate the 600-cell.

void gen600() {

  ld coo[4];
  
  int tab[4];
  
  int tabs[4];
  
  int sig[4];
  
  int a, b, c, q, inv;
  
  /// coordinates taken from Wikipedia
  
  SIGN(tab[0]) SIGN(tab[1]) SIGN(tab[2]) SIGN(tab[3]) {
    vec4 v = zero4;
    for(int i=0; i<4; i++) v[i] = tab[i] / 2.0;
    cell.push_back(v);
    }
  
  REP(i,4) SIGN(q) {
    vec4 v = zero4;
    v[i]=q;
    cell.push_back(v);
    }
  
  coo[0] = 1;
  coo[1] = phi;
  coo[2] = 1/phi;
  coo[3] = 0;
  
  // all permutations
  
  REPD(tab[0], 4) REPD(tab[1], 4) REPD(tab[2], 4) REPD(tab[3], 4) {
    REP(i, 4) REP(j, i)
      if(tab[i] == tab[j]) goto again;

    // check the permutation's sign
    REP(i, 4) tabs[i] = tab[i];
    inv = 0;

    REP(i, 4) while(tabs[i] != i) {
      swap(tabs[i], tabs[tabs[i]]);
      inv++;
      }
    
    if(inv&1) goto again;
    
    // 8 vertices for each permutation
    
    SIGN(sig[0]) SIGN(sig[1]) SIGN(sig[2]) {
      vec4 v = zero4;
      REP(i,4) v[i] = sig[tab[i]] * coo[tab[i]]/2;
      cell.push_back(v);
      }
      
    again: ;
    }
  
  // we add edges between vertices which are close to each other
  // ((specifically in distance 1/phi/phi)
  
  REP(i,vert) REP(j,vert) edge[i][j] = (i != j) && norm(cell[i]-cell[j]) < .4;
  
  // rotate the 600-cell so that it is centered on a 3-simplex
  
  ld angle = 0.2;
  ld maxt = 1;

  while(angle > 1e-30) {
  
    bool changed = false;

    ld sina = sin(angle);
    ld cosa = cos(angle);

    SIGZ(a) SIGZ(b) SIGZ(c) {

      cell2 = cell;
    
      ld curt = 0;

      REP(i,vert) {
        if(a) roll(cell2[i][0], cell2[i][3], sina*a, cosa);
        if(b) roll(cell2[i][1], cell2[i][3], sina*b, cosa);
        if(c) roll(cell2[i][2], cell2[i][3], sina*c, cosa);
        
        if(cell2[i][3] > curt) curt = cell2[i][3];
        }
    
      if(curt < maxt) maxt = curt, cell = cell2, changed = true;
      }
    
    if(!changed) angle /= 2;
    }
  
  /// classify vertices into types
  
  int count = 0;
  REP(i,vert) if(cell2[i][3] > maxt-1e-4) {
    count++;
    vtype[i] = 0;
    } else vtype[i] = 4;
  
  REP(i,vert) if(vtype[i]) REP(j,vert) if(edge[i][j] && !vtype[j]) vtype[i]--;
  
  REP(i,vert) if(vtype[i] == 4) {
    int ct = 0;
    REP(j,vert) if(edge[i][j] && vtype[j] < 4) ct++;

    if(ct == 3) vtype[i] = 5;
    if(ct == 1) vtype[i] = 6;
    if(ct == 0) vtype[i] = 7;
    }
  
  REP(i,vert) if(vtype[i] == 7) {
    int ct = 0;
    REP(j,vert) if(edge[i][j] && vtype[j] < 7) ct++;
    
    if(ct == 5) vtype[i] = 8;
    if(ct == 4) vtype[i] = 9;
    if(ct == 1) vtype[i] = 10;
    if(ct == 0) vtype[i] = 11;    
    }
  
  // generate the faces array
  
  nf = 0;

  REPD(tab[0],vert)
  REPD(tab[1],tab[0]) if(edge[tab[0]][tab[1]])
  REPD(tab[2],vert) if(edge[tab[2]][tab[1]]) if(edge[tab[2]][tab[0]])
  REPD(tab[3],tab[2]) if(edge[tab[3]][tab[0]]) if(edge[tab[3]][tab[1]]) if(edge[tab[3]][tab[2]]) {
  
    REP(i,4) faces[nf][i]= tab[i];
    nf++;
    }
  
  printf("3-simplices: %d/6\n", nf);
  
  /// find the indices of special vertices: 'flat', corners, and redtri

  REP(i,vert) if(vtype[i] == 0) flat = i;

  int ic = 0;
  REP(i,vert) if(vtype[i] == 1) corn[ic++] = i;  
  REP(i,4) if(!edge[flat][corn[i]]) swap(corn[i], corn[3]);
  
  ic = 0;
  REP(i,vert) if(vtype[i] == 0)
    REP(j,vert) if(edge[i][j] && vtype[j] == 3) redtri[ic++] = j;
  }

void projectCell(ld mul1) {
  
  vpos.clear();
  
  /// perform the stereographic projection
  
  REP(i,vert) vpos.push_back(project(cell[i]));

  // Shape the exterior layer (vertices of type 1-3) so that it forms
  // a tetrahedron. This is done by forcing a specific value of 
  // the dot product with the neighboring vertex of type 0

  REP(i,vert) if(ir(vtype[i], 1, 3)) {
    REP(j,vert) if(vtype[j] == 0 && edge[i][j]) {
      ld sca = (vpos[i] | vpos[j]);
      
      vpos[i] *= (mul1/sca);
      
      break;
      }
    }
  
  // move the corners (vertices of type 0) so that a cube is formed
  // (for aesthetic reasons)
  
  REP(i,vert) if(!vtype[i]) {
    vpos[i] = zero3;
    REP(j,vert) if(vtype[j] == 1) {
      if(edge[i][j]) vpos[i] += .5 * vpos[j]; else vpos[i] -= .5 * vpos[j];
      }
    }
  }

// algorithms for verification and flattening
//--------------------------------------------

bool runAlgo = false; // is the algorithm running or is it already finished?

bool dodeca  = false; // are we working on triangulating the dodecahedron?

ld height = 1;        // how flattened so far is the tetrahedron

bool prepared = false;// is dodeca prepared?

ld prog = 0;          // progress of morphing for dodeca

int err;              // how many errors have been found by the verifier

vec3 origa[4];        // origa[i] == original position of corn[i]

vec3 rolla[4];        // where corn[i] should be for dodeca

int subIters;         // iterations of the verification

// mark edge as an edge of type 3

void mark3(int& i) {
  if(i == 1) i = 3;
  }

// verify the triangulation; if co>0, tries to fix the errors

void verify(ld co=0) {

  ld ar = 6 * volume(vpos[corn[0]], vpos[corn[1]], vpos[corn[2]], vpos[corn[3]]);

  // This is 6 times the volume of the tetrahedron. We subtract volumes of
  // all the 3-simplices in our triangulation. The result should be 0,
  // unless something very bad happens (a vertex jumps through a face)

  err = 0;
  
  int tab[4];
  
  int ct = 0;
  
  // an array for fixing
  vector<vec3> newpos = vpos;

  // mark all edges as white (correct)
  REP(i,vert) REP(j,vert) if(edge[i][j])
    edge[i][j] = 1;
  
  REP(u,nf) {

    // tab[0..3] are indices of the 3-simplex considered
    REP(i,4) tab[i] = faces[u][i];
    
    // don't care about exterior vertices of type 0
    if(!vtype[tab[0]]) continue;
    if(!vtype[tab[1]]) continue;
    if(!vtype[tab[2]]) continue;
    if(!vtype[tab[3]]) continue;
    
    // test the volume
    ar -= volume(vpos[tab[0]], vpos[tab[1]], vpos[tab[2]], vpos[tab[3]]);
    
    // we check the dihedral angle at tab[0..1]:
    
    // a1 i a2 are respective cross products
    
    vec3 a1 = (vpos[tab[0]] - vpos[tab[1]]) ^ (vpos[tab[0]] - vpos[tab[2]]);
    vec3 a2 = (vpos[tab[0]] - vpos[tab[1]]) ^ (vpos[tab[0]] - vpos[tab[3]]);
    
    // the dot product of a1 and a2 should be positive
    
    if((a1|a2) <= 1e-6) {

      // mark the simplex as having an error
      err++;
      
      edge[tab[0]][tab[1]] = edge[tab[1]][tab[0]] = 2;
      
      mark3(edge[tab[0]][tab[2]]);
      mark3(edge[tab[0]][tab[3]]);
      mark3(edge[tab[1]][tab[2]]);
      mark3(edge[tab[1]][tab[3]]);
      
      mark3(edge[tab[2]][tab[0]]);
      mark3(edge[tab[3]][tab[0]]);
      mark3(edge[tab[2]][tab[1]]);
      mark3(edge[tab[3]][tab[1]]);
      
      // try to fix it in newpos

      vec3 a0 = vpos[tab[0]], a1 = vpos[tab[1]], a2 = vpos[tab[2]], a3 = vpos[tab[3]];
      
      a1 -= a0; a2 -= a0; a3 -= a0;
      
      a2 -= a1 * (a2|a1) / (a1|a1);

      a3 -= a1 * (a3|a1) / (a1|a1);
      
      a0 = (a2+a3);
      
      a0 /= sqrt(norm(a0));
      
      a0 *= co;
      
      // a0 *= co/10;
      
      newpos[tab[0]] -= a0;
      newpos[tab[1]] -= a0;
      newpos[tab[2]] += a0;
      newpos[tab[3]] += a0;      
      }
    
    }

  // if it exploded, stop the algorithm!
  if(ar<-.05) {
    printf("Explosion!\n");
    runAlgo = false;
    }
  
  // don't move vertices of types 0-2 (or 0-1, for dodeca)
  REP(i,vert) if(vtype[i] < (dodeca ? 2 : 3)) newpos[i] = vpos[i];
  
  // in order to keep the construction clean, move vertices of type 6+ together
  // (unless in dodeca mode)
  if(!dodeca) {
    vec3 shift = zero3; int count = 0;
    REP(i,vert) if(vtype[i] > 5) shift += (newpos[i]-vpos[i]), count++;
    shift /= count;
    REP(i,vert) if(vtype[i] > 5) newpos[i] = vpos[i] + shift;
    }

  // restrict the movement of vertices of type 3 (red ones) to their faces
  REP(a,4) {
    vec3 plane = planevia(vpos[redtri[3*a]], vpos[redtri[3*a+1]], vpos[redtri[3*a+2]]);
    
    REP(b,3) {
      vec3& v = newpos[redtri[3*a+b]];
      // ma byc (v|plane) == 1
      
      v -= plane * ((v|plane) - 1) / (plane|plane);
      }
    }
                      
  // restrict the movement of vertices of type 2 (blue ones) to their faces
  // (for dodeca only)

  if(dodeca) REP(i,120) if(vtype[i] == 2) {
    int nei[2], qnei=0;
    REP(j,120) if(vtype[j] == 1 && edge[i][j]) nei[qnei++] = j;
      
    vec3 v1 = newpos[i] - vpos[nei[0]];
    vec3 v2 = vpos[nei[1]] - vpos[nei[0]];
      
    ld co = (v1|v2) / (v2|v2);

    newpos[i] =
      vpos[nei[0]] + (vpos[nei[1]] - vpos[nei[0]]) * co;
    }
  
  vpos = newpos;  
  }

// get the size of the red triangle
float getDist() {
  ld dist1 = abs(vpos[corn[0]] - vpos[corn[1]]);
  ld dist2 = abs(vpos[redtri[9]] - vpos[redtri[10]]);    
  return dist2/dist1;
  }

// perform a bit of flattening
void flatten(ld co) {

  vec3 pf = vpos[flat];
  
  // save the previous positions
  vector<vec3> pr = vpos;
  
  REP(i,vert) {
    if(vtype[i] <= 3)
      // exterior vertices are flattened in the direction of pf...
      vpos[i] = pr[i] - co * pf * ((pr[i] | pf) / (pf | pf));
    else
      // interior vertices are just scaled...
      vpos[i] *= (1-co*.5);
    }
  
  // we have to move vertices of type 2 to their right positions,
  // based on inscribed circles
  
  // lengths of the long and short edges
  ld ledge  = abs(vpos[corn[0]] - vpos[corn[1]]);
  ld shedge = abs(vpos[corn[0]] - vpos[corn[3]]);
  
  // distance from a vertex of type 2 to corn[3] that we want
  ld reqdist = (shedge+shedge-ledge)/2;
  
  vec3 mall = zero3;
  
  REP(i,vert) if(vtype[i] == 2 && edge[i][corn[3]]) {
    vec3& v = vpos[i];
    vec3& co = vpos[corn[3]];
    
    vec3 v2 = co + (v-co) * reqdist / abs(v-co);
    
    mall += (v2-v);
    v = v2;
    }
  
  // we move the interior vertices (but we want to keep them centered at 0,
  // so this is done by moving the exterior vertices in the opposite direction)

  REP(i,vert) if(vtype[i] < 4) vpos[i] -= (mall/3.5);

  height *= (1-co);
  }

// a step of our algorithm: flattening or verification+fixing

void flattenVerify() {
  ld vc, fc;
    
  vc = .001;
  fc = .01 / height;
  // we flatten by 1% each step; this means 50 steps (the right angled
  // tetrahedron is 50% of the regular one)
    
  if(height < .50001) {
    if(!err) runAlgo = false, fc = 0;
    }
    
  if(err) verify(vc);
  else flatten(fc), verify(vc);
    
  printf("%8.5f %3d %12.9f\n", float(height), err, getDist());
  }

// adjust the regular tetrahedron after we know the getDist of
// the right angled tetrahedron that we obtained

void adjustInner(ld goal) {

  ld co = goal / getDist();
  
  for(int a=0; a<4; a++) {
    vec3 v = vpos[redtri[3*a]] + vpos[redtri[3*a+1]] + vpos[redtri[3*a+2]];
    v /= 3;
    
    for(int b=0; b<3; b++) {
      vpos[redtri[3*a+b]] = 
        v + (vpos[redtri[3*a+b]] - v) * co;
      }
    
    }
  
  for(int i=0; i<vert; i++) if(vtype[i] > 3) vpos[i] *= .9;
  }

// tre algorithm for the dodecahedron
//------------------------------------

void normalizeDodeca();
void saveDodeca();

// prepare all

void prepareDodeca() {

  vec3 a[4];
  
  // vertex #0 is the center of dodecahedron
  a[0][2] = 0; a[0][1] = 0; a[0][0] = 0;
  
  // vertices #1 and #3 are vertices of dodecahedron, on the same face
  a[1][2] = 0; a[1][1] = phi; a[1][0] = -1/phi;
  a[3][2] = 0; a[3][1] = phi; a[3][0] = 1/phi;
  
  // vertex #2 is the center of this face
  a[2][2] = (2+phi)/5; a[2][1] = (1+3*phi)/5; a[2][0] = 0;
  
  // we rotate it, and choose the center, so that it agrees with 
  // the normalized regular tetrahedron as well as possible

  vec3 M[3], MB[3];

  REP(x,3) REP(y,3) MB[x][y] = x==y?12:0;

  ld ang = 0.1;
  int it = 0;
  
  ld bqual = 999999;
  
  ld coa,cob,bcoa=1,bcob=1;
  
  while(ang > 1e-8) {
    it++;
    
    ld C = cos(ang);
    ld S = sin(ang);

    ang *= .999;
    REP(z,3) M[z] = MB[z];
    REP(z,3) {
      switch(it % 3) {
        case 0: roll(M[z][1], M[z][2], S, C); break;
        case 1: roll(M[z][1], M[z][2], -S, C); break;
        }
      }
  
    coa=bcoa, cob=bcob;
    
    if((it%10)%3 == 1) coa += ang;
    if((it%10)%3 == 2) coa -= ang;
    if((it%10)/3 == 1) cob += ang;
    if((it%10)/3 == 2) cob -= ang;

    vec3 center = (a[0]*coa + a[1] + a[3] + a[2]*cob) / (2+coa+cob);
    REP(i,4) a[i] = a[i] - center;


    REP(i,4) rolla[i] = a[i][0] * M[0] + a[i][1] * M[1] + a[i][2] * M[2];
    
    ld qual = 0;
    REP(i,4) {
      vec3 v = (vpos[corn[i]] ^ rolla[i]);
      qual += pow(norm(v), 1.5) / pow(norm(vpos[corn[i]]),1);
      }
    
    if(qual < bqual) {
      bqual = qual; bcoa=coa; bcob=cob;
      REP(z,3) MB[z] = M[z];
      }
    }
  
  printf("coa=%f cob=%f\n", float(bcoa), float(bcob));
  REP(u,4) origa[u] = vpos[corn[u]];
  }

void flattenVerifyDodeca() {

  if(!prepared) {
    normalizeDodeca();
    prepareDodeca();
    prepared = true;
    }

  ld vc, fc;
    
  vc = .001;
  fc = .01;

  verify(vc);
  
  if(err != 0) subIters++;
  
  if(err == 0) {
  
    printf("%8.5f %3d\n", float(prog), subIters); subIters = 0;
    
    if(prog >= 1-1e-6) { 
      runAlgo = false; 
      printf("done\n");
      saveDodeca();
      return;
      }

    prog += fc;
    
    ld coef[120];
    
    REP(i,120) if(vtype[i] == 2) {
      int nei[2], qnei=0;
      REP(j,120) if(vtype[j] == 1 && edge[i][j]) nei[qnei++] = j;
      
      vec3 v1 = vpos[i] - vpos[nei[0]];
      vec3 v2 = vpos[nei[1]] - vpos[nei[0]];
      coef[i] = (v1|v2) / (v2|v2);
      }
    
    ld mul = 1;
    vec3 push = zero3;
    
    REP(i,4) vpos[corn[i]] = rolla[i]*mul*prog + origa[i]*(1-prog) + push;
    
    REP(i,120) if(vtype[i] == 2) {
      int nei[2], qnei=0;
      REP(j,120) if(vtype[j] == 1 && edge[i][j]) nei[qnei++] = j;
      vpos[i] =
        vpos[nei[0]] + (vpos[nei[1]] - vpos[nei[0]]) * coef[i];
      }
  
    // restrict the movement of vertices of type 3 (red ones) to their faces
    REP(a,4) {
    
      int nuls[4], qnul=0;
      REP(i,120) if(vtype[i] == 0) nuls[qnul++] = i;
      
      int ind[3], qind=0;
      REP(i,4) if(edge[corn[i]][nuls[a]]) ind[qind++] = corn[i];
      
      vec3 plane = planevia(vpos[ind[0]], vpos[ind[1]], vpos[ind[2]]);
      
      REP(b,120) if(edge[b][nuls[a]]) if(vtype[b] == 3) {
        vec3& v = vpos[b];

        // goal: (v|plane) == 1
        
        v -= plane * ((v|plane) - 1) / (plane|plane);
        }
      }
    }
  }

// save the cube
//---------------

// normalize the regular tetrahedron so that its corners are
// (0,0,1), (0,1,0), (1,0,0) and (1,1,1)

void normalizeRegular() {
  vec3 v = vpos[flat];

  REP(i,vert) vpos[i] -= v;
  
  vec3 w[3];
  REP(i,3) w[i] = vpos[corn[i]];
  
  REP(i,vert) {
    vec3 v = vpos[i];
    REP(j,3) vpos[i][j] = (v | w[j]) / (w[j] | w[j]);
    }
  }

// normalize the orthogonal tetrahedron so that its corners are
// (0,0,0), (0,0,1), (0,1,0), (1,0,0)
void normalizeOrtho() {
  vec3 v = vpos[corn[3]];
  int i, j;
  REP(i,vert) vpos[i] -= v;
  
  vec3 w[3];
  REP(i,3) w[i] = vpos[corn[i]];
  
  REP(i,vert) {
    vec3 v = vpos[i];
    REP(j,3) vpos[i][j] = (v | w[j]) / (w[j] | w[j]);
    }
  }

// normalize the orthogonal tetrahedron for the sake of triangulation
// of dodecahedron

void normalizeDodeca() {
  vec3 v = vpos[flat];

  REP(i,vert) vpos[i] -= v;
  
  vec3 w[3];
  REP(i,3) w[i] = vpos[corn[i]];
  
  REP(i,vert) {
    vec3 v = vpos[i];
    REP(j,3) vpos[i][j] = 2*(v | w[j]) / (w[j] | w[j]) - 1;

    REP(j,3) vpos[i][j] *= 4;
    }
  
  ld C = cos(M_PI/4);
  REP(i,vert) roll(vpos[i][0], vpos[i][2], -C, C);
  }

// flip the tetrahedron in coordinate a
void flipAll(int a) {
  int i;
  REP(i, vert) vpos[i][a] = 1-vpos[i][a];
  }

// give integer values for coordinates of v
void asInts(vec3& v, int& x, int& y, int& z) {
  x = int(v[0] * 60000+.5);
  y = int(v[1] * 60000+.5);
  z = int(v[2] * 60000+.5);
  }

// save the tetrahedron to f
void saveIt(FILE *f) {
  int xi, yi, zi, xj, yj, zj;
  
  REP(i,vert) REP(j,i) if(edge[i][j]) if(vtype[i] && vtype[j]) {
    asInts(vpos[i], xi, yi, zi);
    asInts(vpos[j], xj, yj, zj);
    fprintf(f, "%8d%8d%8d%8d%8d%8d\n", xi,yi,zi, xj,yj,zj);
    }
  
  int q = 0;
  REP(i,vert) if(vtype[i]) q++;  
  }

// generate the complete triangulation of the cube

void generateTriangulation() {
  FILE *f = fopen("cube.dat", "wt");

  // generate the regular tetrahedron
  projectCell(10.4);

  runAlgo = true; dodeca = false;
  
  // flatten the tetrahedron until it is orthogonal
  while(runAlgo) flattenVerify();

  // normalize and save it
  normalizeOrtho();
  saveIt(f);

  // save other three rotated orthogonal tetrahedrons
  flipAll(0);
  flipAll(1);
  saveIt(f);
  flipAll(1);
  flipAll(2);
  saveIt(f);
  flipAll(0);
  flipAll(1);
  saveIt(f);

  // remember dist = red triangle size
  ld dist = getDist();  

  // generate the new regular tetrahedron, with red triangle size == dist
  projectCell(10.4);
  adjustInner(dist);
  verify();
  printf("verification: %s\n", err ? "error" : "OK");
  
  // normalize and save it
  normalizeRegular();
  saveIt(f);

  fclose(f);
  }

// Save the dodeca.
// See loadDodeca for the description of coordinate system used

void saveDodeca() {
  FILE *f = fopen("dodeca.dat", "wt");
  
  vec3 plane[3];
  ld sub[3], div[3];
  
  REP(coord,3) {
    vec3 t0 = rolla[(3+coord) % 4];
    vec3 t1 = rolla[(3+(coord+1)%3) % 4];
    vec3 t2 = rolla[(3+(coord+2)%3) % 4];

    plane[coord] = planevia(rolla[2], t0, t1);
    
    sub[coord] = (rolla[2]|plane[coord]);
    
    div[coord] = ((t2|plane[coord])-sub[coord]);
    }
  
  REP(i,vert) if(vtype[i] != 0) {
    int coo[3];
    REP(coord,3)
      coo[coord] = int( ((vpos[i] | plane[coord]) - sub[coord]) / div[coord] * 600000 );
    
    fprintf(f, "%7d%7d%7d\n", coo[0], coo[1], coo[2]);
    }
  
  fclose(f);
  }

// load the saved dodecahedron
// 116 vertices are listed in the same order as in the paper
// (i.e. the structure of edges is the same)
// rolla[2] is the origin, the axes are rolla[2]-rolla[i] for i = 1,3,0

void loadDodeca() {
  FILE *f = fopen("dodeca.dat", "rt");
  
  REP(i,vert) if(vtype[i] != 0) {
    int coo[3];
    fscanf(f, "%6d%6d%6d\n", &coo[0], &coo[1], &coo[2]);
    
    vpos[i] = 
      rolla[2] + (
      (rolla[1]-rolla[2]) * coo[0] +
      (rolla[3]-rolla[2]) * coo[1] +
      (rolla[0]-rolla[2]) * coo[2]
      ) / 600000;
    }
  
  fclose(f);
  }

// graphics
//----------

bool hideOk = false;       

int shfrom = 1, shto = 12; // u-i-o-p controlling
                           
int vtc[] = {         // colors for each vertex type
  0xFFFF00, 
  0x00FF00, 0x0000FF, 0xFF0000, 
  0x80FF80, 0x8080FF, 0xFF8080, 
  0xC0FFC0, 0xC0FFFF, 0xC0C0FF, 0xFFC0FF, 0xFFC0C0,
  0xC0C0C0
  };
  
bool hidden(int id) { return vtype[id] < shfrom || vtype[id] > shto; }

// export to SVG format

void saveSvg(int type) {
  f = fopen("cell.svg", "wt");
  fprintf(f, "<svg width=\"%d\" height=\"%d\" xmlns=\"http://www.w3.org/2000/svg\">\n", SX, SY);
  for(int a=0; a<vert; a++)
  for(int b=0; b<a; b++) if(edge[a][b]) {
  
    if(hidden(a)) continue;
    if(hidden(b)) continue;
    
    if(hideOk && edge[a][b] == 1) continue;
    
    vec3 v1 = ekran(vpos[a]);
    vec3 v2 = ekran(vpos[b]);
    
    int vt = max(vtype[a], vtype[b]);
      
    if(type==0)
      addLine(v1, v2, vt <= 3 ? "1px" : "0.25px");
    
    else 
      if(vt <= 3) addLine(v1, v2, vt <= 2 ? "1px" : "0.2px"); // vt <= 3 ? "0.2px" : "0.05px");
    }
  
  fprintf(f, "</svg>\n");
  fclose(f);
  }

// draw and/or animate the algorithm

void frame() {

  if(showit)
  for(int a=0; a<vert; a++)
  for(int b=0; b<a; b++) if(edge[a][b]) {
  
    if(hidden(a)) continue;
    if(hidden(b)) continue;
    
    if(hideOk && edge[a][b] == 1) continue;
    
    vec3 v1 = ekran(vpos[a]);
    vec3 v2 = ekran(vpos[b]);
    
    vec3 av1 = v1, av2 = v2; av1[2] = av2[2] = 0;
    
    ld p = sqrt(norm(av1-av2));
    if(p > 200) p = 200;
    
    for(int i=0; i<p; i++) {
      vec3 v = (v1 * i + v2 * (p-i)) / p;
      
      putpix(int(v[0]), int(v[1]), col(v[2], edge[a][b]));
      }
    }
    
  if(showit)
  for(int a=0; a<vert; a++) {
    if(hidden(a)) continue;
    
    vec3 v1 = ekran(vpos[a]);
    int vx = int(v1[0]), vy = int(v1[1]);
    
    int col = vtc[vtype[a]];
    
    int dx, dy;
    SIGZ(dx) SIGZ(dy) putpix(vx+dx, vy+dy, col);
    
    if(movit[a]) {
      putpix(vx+2, vy+2, 0xFFFF00);
      putpix(vx+2, vy-2, 0xFFFF00);
      putpix(vx-2, vy+2, 0xFFFF00);
      putpix(vx-2, vy-2, 0xFFFF00);
      }    
    }

  if(runAlgo) dodeca ? flattenVerifyDodeca() : flattenVerify();
  }

void applyKey(int e) {

  if(e == '7') vtype[redtri[9]] = 8, vtype[redtri[10]] = 8;
  
  if(e == 'q') stop = true;

  if(e == 'u') shfrom++, printf("%d-%d\n", shfrom, shto);
  if(e == 'i') shfrom--, printf("%d-%d\n", shfrom, shto);
  if(e == 'o') shto++, printf("%d-%d\n", shfrom, shto);
  if(e == 'p') shto--, printf("%d-%d\n", shfrom, shto);

  if(e == 's') hideOk = !hideOk;
  if(e == 'd') showit = !showit;
  if(e == 'r') runAlgo = !runAlgo, dodeca = false;
  if(e == 'c') runAlgo = !runAlgo, dodeca = true;
  
  if(e == '4') printf("%f\n", getDist());

  if(e == 'v') verify(), printf("%d\n", err);
  
  if(e == 't') REP(i,vert) movit[i] = !hidden(i);
  
  if(e == '6') saveSvg(0);

  if(e == '8') saveSvg(1);

  if(e == '.') generateTriangulation();
  
  if(e == 'b') normalizeDodeca(), prepareDodeca(), loadDodeca();
  }

int main() {  
  gen600();
  projectCell(10.4);
  verify();
  graphic("triangulation generator");  
  }
