#include "vector.cpp"

int vert, edg, nf;   // the number of vertices, edges, and faces
int maxco;           // max coordinate

vector<ivec> v;      // vertices (as integers)
vector<vec3> vpos;   // the coordinates of vertices, for drawing

int edge[600][600];  // edge[i][j] > 0 iff cell[i]-cell[j] is an edge;
                     // the specific value denotes the verification status
                     // for this edge (1-white, 2-red, 3-green)


int faces[3000*24][4];// faces (as in cell.cpp)

int err;             // error count

#include "graph.cpp"

// load the triangulation data
//-----------

// what's the number of this vertex? (if we see it for the first time,
// add it to v)

int vertexId(ivec w) {
  for(int i=0; i<vert; i++) if(v[i] == w) return i;
  v.push_back(w);
  return vert++;
  }

void load() {
  vert = edg = nf = 0;
  
  FILE *f = fopen("cube.dat", "rt");
  
  ivec a, b;
    
  while(true) {
    if(fscanf(f, "%d%d%d%d%d%d", &a[0], &a[1], &a[2], &b[0], &b[1], &b[2]) < 0)
      break;
    
    edge[vertexId(a)][vertexId(b)] = true;
    edge[vertexId(b)][vertexId(a)] = true;
    
    edg++;
    }
  
  printf("done\n");
  
  {
  int a,b,c,d;
    
  REP(a,vert) REP(b,vert) if(edge[a][b])
  REP(c,vert) if(edge[a][c]) if(edge[b][c])
  REP(d,vert) if(edge[a][d]) if(edge[b][d]) if(edge[c][d]) {
    faces[nf][0] = a;
    faces[nf][1] = b;
    faces[nf][2] = c;
    faces[nf][3] = d;
    nf++;
    }
  
  maxco = 0;
  REP(a,vert) if(v[a][0] > maxco) maxco = v[a][0];

  printf("vertices   %d\n", vert);
  printf("edges      %d\n", edg);
  printf("3d faces   %d .%d\n", nf/24, nf%24);
  printf("coordinate %d\n", maxco);
  
  vpos.resize(vert);
  REP(a,vert) REP(b,3) vpos[a][b] = 4 * ((v[a][b] - maxco/2.) / maxco);
  }
  }

// verification
//-------------

// This is done as in cell.cpp.

void mark3(int& i) {
  if(i == 1) i = 3;
  }

void bug() {
  printf("overflow!\n");
  exit(1);
  }

// pip: positive inner product (check for overflows)

ll tmul(ll a, ll b) {
  if(a == 0) return 0;
  ll r = a*b;
  if(r/a != b) bug();
  return r;
  }

bool pip(lvec a, lvec b) {
  ll ab0 = tmul(a[0], b[0]);
  ll ab1 = tmul(a[1], b[1]);
  ll ab2 = tmul(a[2], b[2]);
  
  ab0 /= 3; ab1 /= 3; ab2 /= 3;
  
  return (ab0+ab1+ab2) < 3;
  }

void verify(ld co=0) {
                      
  ll vol = 0;

  err = 0;
  
  int tab[4];
  
  REP(i,vert) REP(j,vert) if(edge[i][j]) edge[i][j] = 1;
  
  REP(u,nf) {
    REP(i,4) tab[i] = faces[u][i];
    
    lvec ta[4];
    REP(i,4) ta[i] = asLL(v[tab[i]]);

    vol += volume(ta[0], ta[1], ta[2], ta[3]);
    
    lvec a1 = (ta[0] - ta[1]) ^ (ta[0] - ta[2]);
    lvec a2 = (ta[0] - ta[1]) ^ (ta[0] - ta[3]);
    
    if(pip(a1, a2)) {
      err++;
  
      edge[tab[0]][tab[1]] = edge[tab[1]][tab[0]] = 2;
      
      mark3(edge[tab[0]][tab[2]]);
      mark3(edge[tab[0]][tab[3]]);
      mark3(edge[tab[1]][tab[2]]);
      mark3(edge[tab[1]][tab[3]]);
      
      mark3(edge[tab[2]][tab[0]]);
      mark3(edge[tab[3]][tab[0]]);
      mark3(edge[tab[2]][tab[1]]);
      mark3(edge[tab[3]][tab[1]]);
      }
    
    }

  ll good = 144;
  good *= maxco;
  good *= maxco;
  good *= maxco;
  
  if(good != vol) printf("Illegal volume\n", err++);

  printf("errors: %d\n", err);
  }
  
void saveSvg(const char *fname, int type) {
  f = fopen(fname, "wt");
  fprintf(f, "<svg width=\"%d\" height=\"%d\" xmlns=\"http://www.w3.org/2000/svg\">\n", SX, SY);
  for(int a=0; a<vert; a++)
  for(int b=0; b<a; b++) if(edge[a][b]) {
  
    vec3 v1 = ekran(vpos[a]);
    vec3 v2 = ekran(vpos[b]);
    
    bool small = 
      (v[a][0]%30000)+(v[a][1]%30000)+(v[a][2]%30000)+
      (v[b][0]%30000)+(v[b][1]%30000)+(v[b][2]%30000);
    
    for(int i=0; i<3; i++) {
      int j = (i+1) % 3;
    
      if(v[a][i] == v[b][i] && v[a][j] == v[b][j] && (v[a][i] == 0 || v[a][i] == 60000)) small = false;
      }    
    
    if(type) {
      if(!small) addLine(v1, v2, "1px");
      }
    else {
      vec3 v = v1-v2;
      char buf[20];
      sprintf(buf, "%.2fpx", small ? .1 : 1);
      addLine(v1, v2, buf);
      }
    }
  
  fprintf(f, "</svg>\n");
  fclose(f);
  }

void applyKey(int e) {

  if(e == 'q') stop = true;

  if(e == 'v') verify(), printf("%d\n", err);

  if(e == '6') saveSvg("cube.svg", 0);
  if(e == '8') saveSvg("skel.svg", 1);
  }

void frame() {
  for(int a=0; a<vert; a++)
  for(int b=0; b<a; b++) if(edge[a][b]) {
  
    vec3 v1 = ekran(vpos[a]);
    vec3 v2 = ekran(vpos[b]);
    
    vec3 av1 = v1, av2 = v2; av1[2] = av2[2] = 0;
    
    ld p = sqrt(norm(av1-av2));
    if(p > 200) p = 200;
    
    for(int i=0; i<p; i++) {
      vec3 v = (v1 * i + v2 * (p-i)) / p;
      
      putpix(int(v[0]), int(v[1]), col(v[2], edge[a][b]));
      }
    }
  
  for(int a=0; a<vert; a++) {
    
    vec3 v1 = ekran(vpos[a]);
    int vx = int(v1[0]), vy = int(v1[1]);
    
    int col = 0xFFFF00;
    
    int dx, dy;
    SIGZ(dx) SIGZ(dy) putpix(vx+dx, vy+dy, col);
    
    }
  }

int main() {
  
  load();
  verify();
  graphic("acute triangulation of a cube");  
  }
